// ▼リスト1　スクリプトの構成
"use strict";// strictモード

// 【リスト1-1】クラスの定義
class Character {
    constructor(name, maxHp, maxMp, power, aa) {
        this.name = name;   // 名前
        this.maxHp = maxHp; // 最大HP
        this.maxMp = maxMp; // 最大MP
        this.power = power; // 攻撃力
        this.aa = aa;       // アスキーアート
    }
}

// 【リスト1-2】変数、定数の宣言
// ▼リスト4　キャラクター番号の宣言
const character = {
    player: 0,  // プレイヤー
    slime: 1,   // スライム
    
    // 【リスト4-1】追加キャラクター
    boss: 2,

    max: 3	    // ←【リスト4-2】キャラクターの数
};

// ▼リスト5　キャラクターデータの宣言
let characters = [
    new Character('ゆうしゃ', 15, 10, 3),
    new Character('スライム', 5, 0, 2,
            '　　人<br>'
        +   '（＠ｖ＠）'),
    
    // 【リスト5-1】追加キャラクター
    new Character('ラスボス', 30, 0, 5,
            '　　Ａ　Ａ<br>'
        +   '　　＠＠＠<br>'
        +   'ψ（▼皿▼）ψ')
];

let player;
let enemy;

const phase = {
    encount: 0, // 遭遇フェーズ
    command: 1, // コマンドフェーズ
    action: 2,  // 行動フェーズ
    max: 3
};

let currentPhase;

const command = {
    fight: 0,	// たたかう
    spell: 1,   // じゅもん
    run: 2,     // にげる
    max: 3
};

const commandNames = [
    'たたかう',
    'じゅもん',
    'にげる'
];

currentPhase = phase.command;

let currentCharacter;

const magicCost = 3;

// 【リスト1-3】関数の定義
// ▼リスト2　初期化関数
function init() {
    // 【リスト2-1】キャラクターの初期化
    // ▼リスト6　キャラクターの参照の設定
    player = characters[character.player];
    enemy = characters[character.boss]; // 【リスト6-1】敵の設定
    
    player.target = enemy;
    enemy.target = player;

    player.hp = player.maxHp;
    enemy.hp = enemy.maxHp;

    player.mp = player.maxMp;
    enemy.mp = enemy.maxMp;

    player.command = command.fight;
    enemy.command = command.fight;

    player.damage = player.target.power;
    enemy.damage = enemy.target.power;

    player.commandResult = false;
    enemy.commandResult = false;

    // 【リスト2-2】進行データの初期化
    currentPhase = phase.encount;

    currentCharacter = player;

    // 【リスト2-3】キー入力イベントハンドラーの登録
    window.onkeydown = onKeyDown;

    // 【リスト2-4】描画処理
    draw();
}

// ▼リスト3　描画関数
function draw() {
    let html = '';
    // 【リスト3-1】基本情報の描画
    html += `${player.name}<br>
        ＨＰ：${player.hp}／${player.maxHp}<br>
        ＭＰ：${player.mp}／${player.maxMp}<br>
        こうげきりょく：${player.power}<br>
        <br>`;
    
    html += `${enemy.aa}<br>
        <br>
        （ＨＰ：${enemy.hp}／${enemy.maxHp}）<br>`;

    // 【リスト3-2】フェーズによる分岐
    // ▼リスト8　フェーズによる描画処理の分岐
    html += '<br>';
    switch (currentPhase) {
        case phase.encount:
            // 【リスト8-1】遭遇フェーズ
            html += `${enemy.name}が　あらわれた！<br>`;
            break;
        case phase.command:
            // 【リスト8-2】コマンドフェーズ
            // ▼リスト11　コマンドの一覧描画
            for (let i = 0; i < command.max; i++){
                // 【リスト11-1】カーソルの描画)
                html += (i === currentCharacter.command)
                    ? '＞'
                    : '　';
                
                html += `${commandNames[i]}<br>`;
            }
            break;
        case phase.action:
            // 【リスト8-3】行動フェーズ
            // ▼リスト13　コマンドによる描画処理の分岐
            switch (currentCharacter.command) {
                case command.fight:
                    // 【リスト13-1】「たたかう」
                    html += currentCharacter.name
                        + 'の　こうげき！<br>'
                        + currentCharacter.target.name
                        + 'に　'
                        + currentCharacter.target.damage
                        + 'のダメージ！<br>';
                    
                    if (currentCharacter.target.hp <= 0)
                        html += currentCharacter.target.name
                            + 'を　たおした！<br>';
                    
                    break;
                case command.spell:
                    // 【リスト13-2】「じゅもん」
                    // ▼リスト17　「じゅもん」コマンドのメッセージ処理
                    if (currentCharacter.commandResult) {   // ←【リスト17-1】MPが足りているかのチェック
                        html += currentCharacter.name
                            + `は　ヒールを　となえた！<br>`
                            + currentCharacter.name
                            + `の　きずが　かいふくした！<br>`;
                    }
                    else
                        html += 'ＭＰが　たりない！<br>';   // ←【リスト17-2】MPが足りない場合
                    
                    break;
                case command.run:
                    // 【リスト13-3】「にげる」
                    html += currentCharacter.name
                        + 'は　にげだした！<br>';
                    if (!currentCharacter.commandResult)
                        html += 'しかし　まわりこまれてしまった！<br>';
                    break;
            }
            break;
    }

    // 【リスト3-3】操作方法の描画
    // ▼リスト7　操作方法の描画
    html += '<br>';
    if (currentPhase===phase.command) {    // ←【リスト7-1】コマンドフェーズ判定
        // 【リスト7-2】コマンドの操作方法表示
        html += `[w,s]:カーソル移動<br>
            [その他のキー]:コマンド決定`;
    }
    else
        html += '何かキーを押して下さい。';

    // 【リスト3-4】に追加
    let div = document.querySelector('div');
    div.innerHTML = html;
}

// ▼リスト9　キー入力イベントハンドラ
function onKeyDown(e) {
    // 【リスト9-1】戦闘を終了する処理
    if (
        (player.hp <= 0)
        || (enemy.hp <= 0)) {
        init();
        return;
    }

    if ((currentCharacter.command
        === command.run)
        && currentCharacter.commandResult) {
        init();
        return;
    }

    // 【リスト9-2】フェーズによる分岐
    // ▼リスト10　キー入力時のフェーズによる分岐
    switch (currentPhase) {
        case phase.encount:
            // 【リスト10-1】遭遇フェーズ
            currentPhase = phase.command;
            break;
        case phase.command:
            // 【リスト10-2】コマンドフェーズ
            // ▼リスト12　コマンドの操作
            switch (e.key) {
                case 'w':
                    currentCharacter.command--;
                    break;
                case 's':
                    currentCharacter.command++;
                    break;
                default:
                    // 【リスト12-1】ターンの初期化
                    currentPhase = phase.action;

                    // 【リスト12-2】最初の行動
                    action();
                    break;
            }

            if (currentCharacter.command < 0)
                currentCharacter.command += command.max;
            if (currentCharacter.command >= command.max)
                currentCharacter.command -= command.max;
        
            break;
        case phase.action:
            // 【リスト10-3】行動フェーズ
            if (currentCharacter === player)
                currentCharacter = enemy;
            else if (currentCharacter === enemy)
                currentCharacter = undefined;
        
            // ▼リスト16　行動フェーズのキー入力処理
            if (currentCharacter !== undefined)
                action();
            else {
                // 【リスト16-1】両者のターンが終わったら
                currentCharacter = player;
                currentPhase = phase.command;

                player.commandResult = false;
                enemy.commandResult = false;
            }
            break;
    }
    
    draw();
}

// ▼リスト14　行動フェーズのキー入力時の関数
function action() {
    switch (currentCharacter.command) {
        case command.fight:
            // 【リスト14-1】「たたかう」
            // ▼リスト15　「たたかう」コマンドの処理
            // 【リスト15-1】ダメージ計算)
            currentCharacter.target.damage =
                1 + parseInt(Math.random() * currentCharacter.power);
            
            currentCharacter.target.hp -=
                currentCharacter.target.damage;
            
            // 【リスト15-2】ダメージ制限
            if (currentCharacter.target.hp < 0)
                currentCharacter.target.hp = 0;
    
            break;
        case command.spell:
            // 【リスト14-2】「じゅもん」
            // ▼リスト18　「じゅもん」コマンドのキー入力処理
            if (currentCharacter.mp >= magicCost) { // 【リスト18-1】←(d1)MPが足りているかの判定
                currentCharacter.hp = currentCharacter.maxHp;

                // 【リスト18-2】MPが足りていた場合の処理
                currentCharacter.mp -= magicCost;
                currentCharacter.commandResult = true;
            }
            else
                currentCharacter.commandResult = false; // ←【リスト18-3】MPが足りなかったときの処理
            break;
        case command.run:
            // 【リスト14-3】「にげる」
            currentCharacter.commandResult
                = (Math.random() >= 0.5);
            
            break;
    }
}

// 【リスト1-4】関数の呼び出し
init();