// ▼スクリプトの構成
"use strict"; // strictモード

// 【リスト1-1】クラスの定義
class BlockAA{
    constructor(release, hold) {
        this.release = release;
        this.hold = hold;
    }
}

class Vec2 {
    constructor(x, y) {
        this.x = x;
        this.y = y;
    }
}

// ▼リスト7　カーソルクラスの定義
class Cursor {
    constructor(pos) {
        this.pos = pos;

        // 【リスト7-1】ブロックを掴んでいるフラグ
        this.isHolding = false;
    }
}

// 【リスト1-2】変数／定数の宣言
const block = {
    none: 0,        // なし
    circle: 1,      // ○
    triangle: 2,    // △
    square: 3,      // □
    diamond: 4,     // ◇
    star: 5,        // ☆
    max: 6
};

const blockAA = [
    new BlockAA('　', '　'),
    new BlockAA('○', '●'),
    new BlockAA('△', '▲'),
    new BlockAA('□', '■'),
    new BlockAA('◇', '◆'),
    new BlockAA('☆', '★')
];

let field;

const fieldSize = new Vec2(6, 5);

let cursor;

const direction = {
    horizontal: 0,
    vertical: 1,
    max: 2
};

const directions = [
    new Vec2(1, 0), // horizontal
    new Vec2(0, 1)  // vertical
];

let combo;

// 【リスト1-3】関数の定義
// ▼リスト2　初期化関数
function init() {
    // 【リスト2-1】フィールドの生成
    // ▼リスト5　フィールドの生成
    field = [];
    for (let i = 0; i < fieldSize.y; i++) {
        field[i] = [];
        for (let j = 0; j < fieldSize.x; j++) {
            // 【リスト5-1】ブロックの生成
            let newBlock = 1
                + parseInt(
                    Math.random() * (block.max - 1));
            field[i].push(newBlock);
        }
    }

    // 【リスト2-2】その他の初期化
    cursor = new Cursor(new Vec2(0, 0));

    combo = 0;

    // 【リスト2-3】イベントハンドラの登録
    window.onkeydown = onKeyDown;
    window.onkeyup = onKeyUp;

    // 【リスト2-4】描画処理
    draw();
}

// ▼リスト3　描画関数
function draw() {
    let html = '';
    // 【リスト3-1】フィールドの描画
    // ▼リスト6　フィールドの描画
    for (let i = 0; i < field.length; i++) {
        for (let j = 0; j < field[i].length; j++) {
            let aa = blockAA[field[i][j]].release;

            // 【リスト6-1】掴んでいれば、塗り潰し
            if (cursor.isHolding
                && (i === cursor.pos.y)
                && (j === cursor.pos.x))
                aa = blockAA[field[i][j]].hold;

            html += aa;
        }

        // 【リスト6-2】カーソルの行に矢印を描画する
        if (i === cursor.pos.y)
            html += '←';

        html += '<br>';
    }
    // 【リスト6-3】カーソルの列に矢印を描画する
    for (let i = 0; i < field[0].length; i++)
        if (i === cursor.pos.x)
            html += '↑';
        else
            html += '　';
    html += '<br>';

    // 【リスト3-2】操作方法の描画
    // ▼リスト4　操作方法の描画
    html += '<br>';
    if (combo === 0) { // ←【リスト4-1】コンボ中でなければ
        html += `[w, s, a, d]:カーソル移動<br>
            [その他のキー]:ブロックを`;

        // 【リスト4-2】「掴む／離す」を描画
        html += (cursor.isHolding)
            ? '離す'
            : '掴む';
    }
    else
        html += `${combo} コンボ！`;    // 【リスト4-3】コンボ数の描画

    let div = document.querySelector('div');
    div.innerHTML = html;
}

// ▼リスト8　キー押下イベントハンドラ
function onKeyDown(e) {
    // 【リスト8-1】カーソルの移動前の座標を保存
    let lastPos =
        new Vec2(cursor.pos.x, cursor.pos.y);

    // 【リスト8-2】カーソルの操作
    // ▼リスト10　キー押下処理
    switch (e.key) {
        case 'w': cursor.pos.y--; break;
        case 's': cursor.pos.y++; break;
        case 'a': cursor.pos.x--; break;
        case 'd': cursor.pos.x++; break;
        default:
            // 【リスト10-1】ブロックを掴む
            cursor.isHolding = true;
            break;
    }

    // 【リスト8-3】カーソルの移動制限
    if (cursor.pos.x < 0) cursor.pos.x += fieldSize.x;
    if (cursor.pos.x >= fieldSize.x) cursor.pos.x -= fieldSize.x;
    if (cursor.pos.y < 0) cursor.pos.y += fieldSize.y;
    if (cursor.pos.y >= fieldSize.y) cursor.pos.y -= fieldSize.y;

    // 【リスト8-4】ブロックの入れ替え
    if (cursor.isHolding) {
        let temp = field[lastPos.y][lastPos.x];
        field[lastPos.y][lastPos.x] =
            field[cursor.pos.y][cursor.pos.x];
        field[cursor.pos.y][cursor.pos.x] = temp;
    }

    draw();
}


// ▼リスト9　キーを離したイベントハンドラ
function onKeyUp(e) {
    // 【リスト9-1】掴んでいたブロックを離す
    // ▼リスト11　キーを離したときの処理
    switch (e.key) {
        case 'w':
        case 's':
        case 'a':
        case 'd':
            break;
        default:
            cursor.isHolding = false;

            // 【リスト11-1】コンボを開始する
            startAnimation();

            break;
    }

    draw();
}

// ▼リスト12　ブロックを消す関数
function removeBlocks() {
    let removed = false;
    // 【リスト12-1】削除フラグテーブルの生成
    let toRemove = [];
    for (let i = 0; i < field.length; i++) {
        toRemove[i] = [];
        for (let j = 0; j < field[i].length; j++)
            toRemove[i][j] = false;
    }

    // 【リスト12-2】削除判定
    // ▼リスト13　削除するブロックを調べる
    for (let i = 0; i < field.length; i++)
        for (let j = 0; j < field[i].length; j++) {
            // 【リスト13-1】ブロックの有無判定
            if (field[i][j] === block.none)
                continue;

            // 【リスト13-2】縦横がそろっているか判定
            // ▼リスト14　各方向がそろっているかのチェック
            for (let k = 0; k < directions.length; k++) {
                let v = new Vec2(j, i);
                let chain = 1;
                // 【リスト14-1】連結数のカウント
                // ▼リスト15　連結数のカウントループ
                while (true) {
                    // 【リスト15-1】連結数のカウント
                    v = new Vec2(
                        v.x + directions[k].x,
                        v.y + directions[k].y);
                    
                    if ((v.x >= field[i].length)
                        || (v.y >= field.length)
                    )
                        break;
                    
                    if (field[v.y][v.x] === field[i][j])
                        chain++;
                    else
                        break;
                }

                // 【リスト14-2】そろったら削除チェック
                // ▼リスト16　削除チェック
                if (chain >= 3) {
                    removed = true;
                    // 【リスト16-1】消すブロックにチェック
                    v = new Vec2(j, i);
                    for (let l = 0; l < chain; l++) {
                        toRemove[v.y][v.x] = true;
                        v = new Vec2(
                            v.x + directions[k].x,
                            v.y + directions[k].y);
                    }
                }
            }

        }

    // 【リスト12-3】削除実行
    for (let i = 0; i < field.length; i++)
        for (let j = 0; j < field[i].length; j++)
            if (toRemove[i][j])
                field[i][j] = block.none;

    // 【リスト12-4】コンボ数のカウント
    if (removed)
        combo++;

    // 【リスト12-5】再描画
    if (removed)
        draw();

    return removed;
}

// ▼リスト17　ブロックを落下させる関数
function dropBlocks() {
    let dropped = false;
    // 【リスト17-1】ブロックを落下させる
    // ▼リスト18　ブロックを落下させる反復
    for (let i = field.length - 1; i > 0; i--) {
        for (let j = 0; j < field[i].length; j++) {
            // 【リスト18-1】各ブロックを落下させる
            if ((field[i][j] === block.none)
                && (field[i - 1][j] !== block.none)) {
                dropped = true;
                field[i][j] = field[i - 1][j];
                field[i - 1][j] = block.none;
            }
        }
    }
    
    // 【リスト17-2】最上段にブロックを発生させる
    for (let i = 0; i < field[0].length; i++)
        if (field[0][i] === block.none) {
            dropped = true;
            field[0][i] = 1 + parseInt(
                Math.random() * (block.max - 1));
        }

    // 【リスト17-3】再描画
    if (dropped)
        draw();

    return dropped;
}

// ▼リスト19　コンボの開始
function startAnimation() {
    window.onkeydown = null;
    window.onkeyup = null;

    // 【リスト19-1】コンボ数の初期化
    combo = 0;

    updateAnimation();
}

// ▼リスト20　コンボの更新
function updateAnimation() {
    let dropped = true; 
    // 【リスト20-1】ブロックの落下
    dropped = dropBlocks();

    if (!dropped) {
        let removed = true;
        // 【リスト20-2】ブロックの削除
        removed = removeBlocks();

        if (!removed) {
            // 【リスト20-3】コンボ数のリセット
            combo = 0;
            draw();

            // 【リスト20-4】コンボの終了
            endAnimation();

            return;
        }
    }

    // 【リスト20-5】一定時間後に再び更新する
    setTimeout(updateAnimation, 1000);
}

function endAnimation() {
    window.onkeydown = onKeyDown;
    window.onkeyup = onKeyUp;
}

// 【リスト1-4】関数の呼び出し
init();